/* Copyright (C) 1998-1999 Kimmo.palin@helsinki.fi
   License: see COPYING
*/
#include<stdlib.h>
#include<stdio.h>
#include<assert.h>
#include<limits.h>

#include"cpu.h"
#include"cpu_access.h"
#include"mem_access.h" /* for getCoreSize() */

static void *_execing_instr_=NULL,*_full_run_=NULL;


int numOfCPUs=0;

char *print_cpu_stat(cpudev *cpu,char *buffer)
{
  int i;

  sprintf(buffer,"#%d pc=%d sr=%d tr=%d sp=%d fp=%d\n",
	 cpu->CPUid,cpu->pc,cpu->sr,cpu->tr,cpu->regs[SPREG],cpu->regs[FPREG]);
  for(i=0;i<SPREG;i++)
    sprintf(buffer,"%sR%d=%d ",buffer,i,cpu->regs[i]);
  
  return buffer;
}

cpudev *init_cpu()
{
  /* We boot a new CPU with full memory access. The SMP support
     requires something else, like way of identify running CPU 
     i.e. Interrupts */
 
  cpudev *ring;

  ring=(cpudev*)malloc(sizeof(cpudev));
  SET_PRIVLIGED(ring);

  ring->CPUid=numOfCPUs++;
  set_pc(ring,0);
  set_reg(ring,memlimit,SPREG);
  set_reg(ring,memlimit,FPREG);
  set_base(ring,0);
  set_limit(ring,getCoreSize()-1);
  UNSET_PRIVLIGED(ring);

  return ring;

}

/* Timeout function to implement continous running of the process. */
gint process_timeout(gpointer data)
{
  cpudev *ring=(cpudev*)data;
  memunit tmp;

  if(getCoreSize()>get_pc(ring) && 
	(get_tr(ring)!=S_HALT || (get_sr(ring)&STAT_S)==0) && 
	(get_sr(ring)&(STAT_U|STAT_M))==0) {
    tmp=getMem(ring,get_pc(ring));
    set_ir(ring,tmp.flds);
    set_pc(ring,get_pc(ring)+1);

    assert(get_ir(ring).opcode==tmp.flds.opcode);
    exec_instr(ring);
    if(_execing_instr_) {
      _full_run_=ring;
      return FALSE;
    }
  }
  if(get_sr(ring)&STAT_U) {
    fprintf(stderr,"Tuntematon tai virheellinen komento (%d) (pc=%d)\n",
	    get_ir(ring).opcode,get_pc(ring)-1);
    return FALSE;
  }
  if(get_sr(ring)&STAT_S && get_tr(ring)==S_HALT) {
    set_message("Ohjelma pyshtynyt!\n");
    return FALSE;
  }


  return TRUE;
}

/* ring == running CPU */
cpudev *run_cpu(cpudev *ring)  
{
  memunit tmp;

  if(ring==NULL) {
    ring=init_cpu();
  }

  while(getCoreSize()>get_pc(ring) && 
	(get_tr(ring)!=S_HALT || (get_sr(ring)&STAT_S)==0) && 
	(get_sr(ring)&(STAT_U|STAT_M))==0) {
    tmp=getMem(ring,get_pc(ring));
    set_ir(ring,tmp.flds);
    set_pc(ring,get_pc(ring)+1);

    assert(get_ir(ring).opcode==tmp.flds.opcode);
    exec_instr(ring);
    if(_execing_instr_) {
      _full_run_=ring;
      return ring;
    }
  }
  if(get_sr(ring)&STAT_U) {
    fprintf(stderr,"Tuntematon tai virheellinen komento (%d) (pc=%d)\n",
	    get_ir(ring).opcode,get_pc(ring)-1);
  }
  if(get_sr(ring)&STAT_S && get_tr(ring)==S_HALT)
    set_message("Ohjelma pyshtynyt!\n");

  free(ring);

  return NULL;
}

/* Jos huvittaa, tmn voisi jakaa viel osiin CU,MMU,ALU*/
void exec_instr(cpudev *cpu)
{
  memunit tmp;
  opdata instr;
  char buf[10];

  _execing_instr_=cpu;
  instr=get_ir(cpu);

  asmOp(buf,instr.opcode);
  /*printf("(pc=%d) %s\n",get_pc(cpu),buf);
   */
  set_tr(cpu,instr.addr);
  if(instr.ri!=0)
    set_tr(cpu,get_tr(cpu)+get_reg(cpu,instr.ri));

  if((instr.opcode&OPMASK)==JUMP_OPS ||
     instr.opcode==STORE) {
    if(instr.mem==0){
      or_sr(cpu,STAT_U);
      return;
    }
    while(instr.mem>1) {
      tmp=getMem(cpu,get_tr(cpu));
      set_tr(cpu,tmp.bin);
      instr.mem--;
    }
  } else if(instr.opcode!=POP) {
    while(instr.mem>0) {
      tmp=getMem(cpu,get_tr(cpu));
      set_tr(cpu,tmp.bin);
      instr.mem--;
    }
  }

  switch(instr.opcode) {
  case NOP: break;
  case STORE: tmp.bin=get_reg(cpu,instr.rj);
    setMem(cpu,get_tr(cpu),tmp);
    break;
  case LOAD: set_reg(cpu,get_tr(cpu),instr.rj);
    break;
  case IN: if(get_tr(cpu)==D_KBD) {  /* Interrupts suck in TTK91*/
    kbd_input();
  } else or_sr(cpu,STAT_U);
  break;
  case OUT: if(get_tr(cpu)==D_CRT)
    crt_message(get_reg(cpu,instr.rj));
    else or_sr(cpu,STAT_U);
  break;
  case ADD: set_reg(cpu,get_reg(cpu,instr.rj)+get_tr(cpu),instr.rj);
    break;
  case SUB: set_reg(cpu,get_reg(cpu,instr.rj)-get_tr(cpu),instr.rj);
    break;
  case MUL:  set_reg(cpu,get_reg(cpu,instr.rj)*get_tr(cpu),instr.rj);
    break;
  case DIV: set_reg(cpu,get_reg(cpu,instr.rj)/get_tr(cpu),instr.rj);
    break;
  case MOD:  set_reg(cpu,get_reg(cpu,instr.rj)%get_tr(cpu),instr.rj);
    break;
  case AND: set_reg(cpu,get_reg(cpu,instr.rj)&get_tr(cpu),instr.rj);
    break;
  case OR:  set_reg(cpu,get_reg(cpu,instr.rj)|get_tr(cpu),instr.rj);
    break;
  case XOR:  set_reg(cpu,get_reg(cpu,instr.rj)^get_tr(cpu),instr.rj);
    break;
  case SHL: set_reg(cpu,get_reg(cpu,instr.rj)<<get_tr(cpu),instr.rj);
    break;
  case SHR:  set_reg(cpu,get_reg(cpu,instr.rj)>>get_tr(cpu),instr.rj);
    break;
  case COMP: set_sr(cpu,get_sr(cpu)&NULLCOMP);
    if(get_reg(cpu,instr.rj)<get_tr(cpu))
      or_sr(cpu,STAT_L);
    else if(get_reg(cpu,instr.rj)>get_tr(cpu))
      or_sr(cpu,STAT_G);
  else or_sr(cpu,STAT_E);
  break;
  case JUMP: set_pc(cpu,get_tr(cpu));
    break;
  case JNEG: if(get_reg(cpu,instr.rj)<0)
    set_pc(cpu,get_tr(cpu));
  break;
  case JZER: if(get_reg(cpu,instr.rj)==0)
    set_pc(cpu,get_tr(cpu));
  break;
  case JPOS: if(get_reg(cpu,instr.rj)>0)
    set_pc(cpu,get_tr(cpu));
  break;
  case JNNEG: if(get_reg(cpu,instr.rj)>=0)
    set_pc(cpu,get_tr(cpu));
  break;
  case JNZER: if(get_reg(cpu,instr.rj)!=0)
    set_pc(cpu,get_tr(cpu));
  break;
  case JNPOS: if(get_reg(cpu,instr.rj)<=0)
    set_pc(cpu,get_tr(cpu));
  break;
  case JLES: if(get_sr(cpu)&STAT_L)
    set_pc(cpu,get_tr(cpu));
  break;
  case JEQU: if(get_sr(cpu)&STAT_E)
    set_pc(cpu,get_tr(cpu));
  break;
  case JGRE: if(get_sr(cpu)&STAT_G)
    set_pc(cpu,get_tr(cpu));
  break;
  case JNLES: if(get_sr(cpu)&(STAT_E|STAT_G))
    set_pc(cpu,get_tr(cpu));
  break;
  case JNEQU: if(get_sr(cpu)&(STAT_L|STAT_G))
    set_pc(cpu,get_tr(cpu));
  break;
  case JNGRE: if(get_sr(cpu)&(STAT_E|STAT_L))
    set_pc(cpu,get_tr(cpu));
  break;
  case CALL: /*Mikrokooditoteutus
    tmpforexit=(bit32)ttk_tr;
    instr.opcode=PUSH;
    instr.mem=0;
    instr.addr=ttk_pc;
    exec_instr();
    instr.addr=get_reg(cpu,FPREG);
    exec_instr();
    ttk_regs[FPREG]=get_reg(cpu,instr.rj);
    ttk_pc=tmpforexit;*/

    set_reg(cpu,get_reg(cpu,instr.rj)+1,instr.rj);
    tmp.bin=get_pc(cpu);
    setMem(cpu,get_reg(cpu,instr.rj),tmp);

    set_reg(cpu,get_reg(cpu,instr.rj)+1,instr.rj);
    tmp.bin=get_reg(cpu,FPREG);
    setMem(cpu,get_reg(cpu,instr.rj),tmp);

    set_reg(cpu,get_reg(cpu,instr.rj),FPREG);

    set_pc(cpu,get_tr(cpu));
    break;
  case EXIT: /* Tm on hankala koska pc:hen ei saa popattua
    tmpforexit=(bit32)ttk_tr;
    instr.opcode=POP;
    instr.mem=0;
    instr.ri=FPREG;
    exec_instr();

    tmp=getMem(ttk_regs[SPREG]);
    ttk_regs[SPREG]-=(1+tmpforexit);
    ttk_pc=(bit32)tmp.bin;*/

    tmp=getMem(cpu,get_reg(cpu,instr.rj));
    set_reg(cpu,tmp.bin,FPREG);
    set_reg(cpu,get_reg(cpu,instr.rj)-1,instr.rj);

    tmp=getMem(cpu,get_reg(cpu,instr.rj));
    set_pc(cpu,tmp.bin);
    set_reg(cpu,get_reg(cpu,instr.rj)-1-get_tr(cpu),instr.rj);
    break;
  case PUSH: 
    tmp.bin=get_tr(cpu);
    set_reg(cpu,get_reg(cpu,instr.rj)+1,instr.rj);
    setMem(cpu,get_reg(cpu,instr.rj),tmp);
    break;
  case POP:tmp=getMem(cpu,get_reg(cpu,instr.rj));
    set_reg(cpu,tmp.bin,instr.ri);
    set_reg(cpu,get_reg(cpu,instr.rj)-1,instr.rj);
    break;
  case SVC: or_sr(cpu,STAT_S|STAT_P);
    break;
  case IRET:
  default:
    or_sr(cpu,STAT_U);
    break;
  }
  /*  printf("%s\n",print_cpu_stat(cpu,cpustat));*/
  assert(cpu->pc>=0);

  if(instr.opcode!=IN) /* Fiddeling with input */
    _execing_instr_=NULL;

}

void kbd_input_done(int value)
{
  opdata instr=get_ir((cpudev*)_execing_instr_);

  set_reg((cpudev*)_execing_instr_,value,instr.rj);
  _execing_instr_=NULL;

  /*  if(_full_run_) {
    run_cpu((cpudev*)_execing_instr_);
  }*/

}

/*
#define asmOp(code) AsmOps[(code>POP?0:code)]
*/
/*Miksei tm toiminut taulukkona???? */
char *asmOp(char *buffer,bit8 code)
{
  char *tmp;

  switch(code) {
  case NOP: tmp="NOP";break;
  case STORE	: tmp="STORE";break;
  case LOAD	: tmp="LOAD";break;
  case IN		: tmp="IN";break;
  case OUT		: tmp="OUT";break;
  case ADD		: tmp="ADD";break;
  case SUB		: tmp="SUB";break;
  case MUL		: tmp="MUL";break;
  case DIV		: tmp="DIV";break;
  case MOD		: tmp="MOD";break;
  case AND		: tmp="AND";break;
  case OR		: tmp="OR";break;
  case XOR		: tmp="XOR";break;
  case SHL		: tmp="SHL";break;
  case SHR		: tmp="SHR";break;
  case COMP	: tmp="COMP";break;
  case JUMP	: tmp="JUMP";break;
  case JNEG	: tmp="JNEG";break;
  case JZER	: tmp="JZER";break;
  case JPOS	: tmp="JPOS";break;
  case JNNEG	: tmp="JNNEG";break;
  case JNZER	: tmp="JNZER";break;
  case JNPOS	: tmp="JNPOS";break;
  case JLES	: tmp="JLES";break;
  case JEQU	: tmp="JEQU";break;
  case JGRE	: tmp="JGRE";break;
  case JNLES	: tmp="JNLES";break;
  case JNEQU	: tmp="JNEQU";break;
  case JNGRE	: tmp="JNGRE";break;
  case CALL	: tmp="CALL";break;
  case EXIT	: tmp="EXIT";break;
  case PUSH	: tmp="PUSH";break;
  case POP		: tmp="POP";break;
  case SVC		: tmp="SVC";break;
  case IRET	: tmp="IRET";break;
  default: tmp="**TUNTEMATON**"; break;
  }
  if(buffer!=NULL) {
    strcpy(buffer,tmp);
    return buffer;
  }
  return tmp;
}


