/* (c) Juho Snellman (jsnell@lyseo.edu.ouka.fi) 1998.
   Copyright (C) 1998-1999 Kimmo.palin@helsinki.fi
   License: see COPYING
*/

/* This program is under the Gnu General Public License. For more detailed
 * licensing information see the file COPYING included with the 
 * distribution */

/* Major changes by kimmo.palin@helsinki.fi*/

#ifndef _TTKVM_H_
#define _TTKVM_H_

#ifndef VERSION
#define VERSION "0.0.1"
#endif
/* includes */
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>


/* defines */

typedef signed int bit32;
typedef signed short int bit16;
typedef unsigned char bit8;

/*sr:n bittimaskit*/
/*Vertailu*/
/* Nollaa vertailut */
#define NULLCOMP 0xFFFFFFF0


#define STAT_G 1       /* Greater */
#define STAT_E 2       /* Equal */
#define STAT_L 4       /* Less */
/*Poikkeus*/
#define STAT_O 8        /* Overflow */
#define STAT_Z 16       /* Division by zero */
#define STAT_U 32       /* Unknown instruction */
#define STAT_M 64       /* Forbidden memory access */
/*Keskeytys:*/
#define STAT_I 128      /* Interrupt */
#define STAT_S 256      /* Supervisor call */
#define STAT_P 512      /* Privliged mode */
#define STAT_D 1024     /* Interrupts disabled */

typedef struct _opdata {
  bit8 opcode;
  unsigned rj:3;
  unsigned mem:2;
  unsigned ri:3;
  bit16 addr;
} opdata;


/*Ensimminen (staattisesti) varaamaton muistisana*/
extern bit32 memlimit;
/*Ohjelmakoodia seuraava muistisana*/
extern bit32 progcount;

  /* Kskyt */

#define NOOFOPS 0x72

#define OPMASK 0xF0
/*Hallinnoliset*/
#define DATA_OPS  0x00; 

#define NOP		0x00
#define STORE	0x01
#define LOAD	0x02
#define IN		0x03
#define OUT		0x04

/*Aritmeettiset*/
#define ARITH_OPS  0x10

#define ADD		0x11
#define SUB		0x12
#define MUL		0x13
#define DIV		0x14
#define MOD		0x15
#define AND		0x16
#define OR		0x17
#define XOR		0x18
#define SHL		0x19
#define SHR		0x1A
#define COMP	0x1F

/* Hypyt*/
#define JUMP_OPS 0x20

#define JUMP	0x20
#define JNEG	0x21
#define JZER	0x22
#define JPOS	0x23
#define JNNEG	0x24
#define JNZER	0x25
#define JNPOS	0x26
#define JLES	0x27
#define JEQU	0x28
#define JGRE	0x29
#define JNLES	0x2A
#define JNEQU	0x2B
#define JNGRE	0x2C
#define CALL	0x2D

/* Monimutkaiset */
#define CMPLX_OPS 0x40

#define PUSH	0x43
#define POP	0x44
#define EXIT	0x4E

/*Palvelut*/
#define SERV_OPS 0x80

#define SVC	0x80
#define IRET	0x81         /* Miks tm on ? */



  /* devicet */

#define D_CRT 8
#define D_KBD 7

  /* syscallit */

#define S_HALT 11

void err_message(char *message);
void set_message(char *message);

/* Verbose errormessages! */
#define YYERROR_VERBOSE 1


#endif





